/*
 * WPA Supplicant - driver interaction with Linux ipw2100 driver
 * Copyright (c) 2004 Lubomir Gelo <lgelo@cnc.sk>
 * Copyright (c) 2003-2004, Jouni Malinen <jkmaline@cc.hut.fi>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * Alternatively, this software may be distributed under the terms of BSD
 * license.
 *
 * See README and COPYING for more details.
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/ioctl.h>
#include <errno.h>

#include "wireless_copy.h"
#include "common.h"
#include "driver.h"
#include "l2_packet.h"
#include "driver_wext.h"
#include "wpa_supplicant.h"

/* following definitions must be kept in sync with ipw2100_main.c */

#define IPW2100_IOCTL_WPA_SUPPLICANT		SIOCIWFIRSTPRIV+30

#define IPW2100_CMD_SET_WPA_PARAM		1
#define	IPW2100_CMD_SET_WPA_IE			2
#define IPW2100_CMD_SET_ENCRYPTION		3
#define IPW2100_CMD_MLME			4

#define IPW2100_PARAM_WPA_ENABLED		1
#define IPW2100_PARAM_TKIP_COUNTERMEASURES	2
#define IPW2100_PARAM_DROP_UNENCRYPTED		3
#define IPW2100_PARAM_PRIVACY_INVOKED		4
#define IPW2100_PARAM_AUTH_ALGS			5
#define IPW2100_PARAM_IEEE_802_1X		6

#define IPW2100_MLME_STA_DEAUTH			1
#define IPW2100_MLME_STA_DISASSOC		2

#define IPW2100_CRYPT_ERR_UNKNOWN_ALG		2
#define IPW2100_CRYPT_ERR_UNKNOWN_ADDR		3
#define IPW2100_CRYPT_ERR_CRYPT_INIT_FAILED	4
#define IPW2100_CRYPT_ERR_KEY_SET_FAILED	5
#define IPW2100_CRYPT_ERR_TX_KEY_SET_FAILED	6
#define IPW2100_CRYPT_ERR_CARD_CONF_FAILED	7

#define	IPW2100_CRYPT_ALG_NAME_LEN		16

struct ipw2100_param {
	u32 cmd;
	u8 sta_addr[ETH_ALEN];
        union {
		struct {
			u8 name;
			u32 value;
		} wpa_param;
		struct {
			u32 len;
			u8 *data;
		} wpa_ie;
	        struct{
			int command;
    			int reason_code;
		} mlme;
		struct {
			u8 alg[IPW2100_CRYPT_ALG_NAME_LEN];
			u8 set_tx;
			u32 err;
			u8 idx;
			u8 seq[8];
			u16 key_len;
			u8 key[0];
		} crypt;

	} u;
};

/* end of ipw2100_main.c code */

static int ipw2100_ioctl(const char *dev, struct ipw2100_param *param,
			 int len, int show_err)
{
	int s;
	struct iwreq iwr;

	s = socket(PF_INET, SOCK_DGRAM, 0);
	if (s < 0) {
		perror("socket");
		return -1;
	}

	memset(&iwr, 0, sizeof(iwr));
	strncpy(iwr.ifr_name, dev, IFNAMSIZ);
	iwr.u.data.pointer = (caddr_t) param;
	iwr.u.data.length = len;

	if (ioctl(s, IPW2100_IOCTL_WPA_SUPPLICANT, &iwr) < 0) {
		int ret;
		close(s);
		ret = errno;
		if (show_err) 
			perror("ioctl[IPW2100_IOCTL_WPA_SUPPLICANT]");
		return ret;
	}
	close(s);

	return 0;
}

static void ipw2100_show_set_key_error(struct ipw2100_param *param)
{
	switch (param->u.crypt.err) {
	case IPW2100_CRYPT_ERR_UNKNOWN_ALG:
		wpa_printf(MSG_INFO, "Unknown algorithm '%s'.",
			   param->u.crypt.alg);
		wpa_printf(MSG_INFO, "You may need to load kernel module to "
			   "register that algorithm.");
		wpa_printf(MSG_INFO, "E.g., 'modprobe ieee80211_crypt_wep' for "
			   "WEP.");
		break;
	case IPW2100_CRYPT_ERR_UNKNOWN_ADDR:
		wpa_printf(MSG_INFO, "Unknown address " MACSTR ".",
			   MAC2STR(param->sta_addr));
		break;
	case IPW2100_CRYPT_ERR_CRYPT_INIT_FAILED:
		wpa_printf(MSG_INFO, "Crypt algorithm initialization failed.");
		break;
	case IPW2100_CRYPT_ERR_KEY_SET_FAILED:
		wpa_printf(MSG_INFO, "Key setting failed.");
		break;
	case IPW2100_CRYPT_ERR_TX_KEY_SET_FAILED:
		wpa_printf(MSG_INFO, "TX key index setting failed.");
		break;
	case IPW2100_CRYPT_ERR_CARD_CONF_FAILED:
		wpa_printf(MSG_INFO, "Card configuration failed.");
		break;
	}
}

static int ipw2100_set_wpa_ie(const char *ifname, const char *wpa_ie, 
			      size_t wpa_ie_len)
{
	struct ipw2100_param *param;
	int ret;
	size_t blen = sizeof(*param) + wpa_ie_len;

	param = (struct ipw2100_param *) malloc(blen);
	if (param == NULL)
		return -1;

	memset(param, 0, blen);
	param->cmd = IPW2100_CMD_SET_WPA_IE;
	param->u.wpa_ie.len = wpa_ie_len;

	if (wpa_ie_len){
	    param->u.wpa_ie.data = (u8 *)(param+1);
	    memcpy(param->u.wpa_ie.data, wpa_ie, wpa_ie_len);
	} else {
	    param->u.wpa_ie.data = NULL;
	}
	
	ret = ipw2100_ioctl(ifname, param, blen, 1);

	free(param);
	return ret;
}

static int ipw2100_set_wpa_param(const char *ifname, u8 name, u32 value)
{
	struct ipw2100_param param;

	memset(&param, 0, sizeof(param));
	param.cmd = IPW2100_CMD_SET_WPA_PARAM;
	param.u.wpa_param.name = name;
	param.u.wpa_param.value = value;

	return ipw2100_ioctl(ifname, &param, sizeof(param), 1);
}

static int ipw2100_mlme(const char *ifname, u8 *addr, int cmd, int reason)
{
	struct ipw2100_param param;

	memset(&param, 0, sizeof(param));
	memcpy(param.sta_addr, addr, ETH_ALEN);	
	param.cmd = IPW2100_CMD_MLME;
	param.u.mlme.command = cmd;
	param.u.mlme.reason_code = reason;

	return ipw2100_ioctl(ifname, &param, sizeof(param), 1);
}


static int wpa_driver_ipw2100_set_wpa(const char *ifname, int enabled)
{
	int ret = 0;

	wpa_printf(MSG_DEBUG, "%s: enabled=%d", __FUNCTION__, enabled);

	if (!enabled && ipw2100_set_wpa_ie(ifname, NULL, 0) < 0)
		ret = -1;

	if (ipw2100_set_wpa_param(ifname, IPW2100_PARAM_WPA_ENABLED, enabled) < 0)
		ret = -1;

	return ret;
}

static int wpa_driver_ipw2100_set_key(const char *ifname, wpa_alg alg,
				      unsigned char *addr, int key_idx,
				      int set_tx, u8 *seq, size_t seq_len,
				      u8 *key, size_t key_len)
{
	struct ipw2100_param *param;
	u8 *buf;
	size_t blen;
	int ret = 0;
	char *alg_name;

	switch (alg) {
	case WPA_ALG_NONE:
		alg_name = "none";
		break;
	case WPA_ALG_WEP:
		alg_name = "WEP";
		break;
	case WPA_ALG_TKIP:
		alg_name = "TKIP";
		break;
	case WPA_ALG_CCMP:
		alg_name = "CCMP";
		break;
	default:
		return -1;
	}

	wpa_printf(MSG_DEBUG, "%s: alg=%s key_idx=%d set_tx=%d seq_len=%d "
		   "key_len=%d", __FUNCTION__, alg_name, key_idx, set_tx,
		   seq_len, key_len);

	if (seq_len > 8)
		return -2;

	blen = sizeof(*param) + key_len;
	buf = malloc(blen);
	if (buf == NULL)
		return -1;
	memset(buf, 0, blen);

	param = (struct ipw2100_param *) buf;
	param->cmd = IPW2100_CMD_SET_ENCRYPTION;
	memset(param->sta_addr, 0xff, ETH_ALEN);
	strncpy(param->u.crypt.alg, alg_name, IPW2100_CRYPT_ALG_NAME_LEN);
	param->u.crypt.set_tx = set_tx ? 1 : 0;
	param->u.crypt.idx = key_idx;
	memcpy(param->u.crypt.seq, seq, seq_len);
	param->u.crypt.key_len = key_len;
	memcpy((u8 *) (param + 1), key, key_len);

	if (ipw2100_ioctl(ifname, param, blen, 1)) {
		wpa_printf(MSG_WARNING, "Failed to set encryption.");
		ipw2100_show_set_key_error(param);
		ret = -1;
	}
	free(buf);

	return ret;
}

static int wpa_driver_ipw2100_set_countermeasures(const char *ifname,
						 int enabled)
{
	wpa_printf(MSG_DEBUG, "%s: enabled=%d", __FUNCTION__, enabled);
	return ipw2100_set_wpa_param(ifname, IPW2100_PARAM_TKIP_COUNTERMEASURES, enabled);

}


static int wpa_driver_ipw2100_set_drop_unencrypted(const char *ifname,
						  int enabled)
{
	wpa_printf(MSG_DEBUG, "%s: enabled=%d", __FUNCTION__, enabled);
	return ipw2100_set_wpa_param(ifname, IPW2100_PARAM_DROP_UNENCRYPTED, enabled);
}


static int wpa_driver_ipw2100_deauthenticate(const char *ifname, u8 *addr,
					     int reason_code)
{
	
	return ipw2100_mlme(ifname, addr, IPW2100_MLME_STA_DEAUTH, reason_code);;
}


static int wpa_driver_ipw2100_disassociate(const char *ifname, u8 *addr,
					   int reason_code)
{
	return ipw2100_mlme(ifname, addr, IPW2100_MLME_STA_DISASSOC, reason_code);;
}


static int wpa_driver_ipw2100_associate(const char *ifname, const char *bssid,
					const char *ssid, size_t ssid_len,
					int freq,
					const char *wpa_ie, size_t wpa_ie_len,
					wpa_cipher pairwise_suite,
					wpa_cipher group_suite,
					wpa_key_mgmt key_mgmt_suite)
{
	int ret = 0;
	int unencrypted_eapol;

	if (ipw2100_set_wpa_ie(ifname, wpa_ie, wpa_ie_len) < 0)
		ret = -1;
	if (wpa_driver_wext_set_ssid(ifname, ssid, ssid_len) < 0)
		ret = -1;
	if (wpa_driver_wext_set_bssid(ifname, bssid) < 0)
		ret = -1;

	if (key_mgmt_suite == KEY_MGMT_802_1X ||
	    key_mgmt_suite == KEY_MGMT_PSK)
		unencrypted_eapol = 0;
	else
		unencrypted_eapol = 1;
	
	if (ipw2100_set_wpa_param(ifname, IPW2100_PARAM_IEEE_802_1X,
				  unencrypted_eapol) < 0) {
		wpa_printf(MSG_DEBUG, "ipw2100: Failed to configure "
			   "ieee_802_1x param");
	}

	return ret;
}


static int wpa_driver_ipw2100_set_auth_alg(const char *ifname, int auth_alg)
{
	int algs = 0;

	if (auth_alg & AUTH_ALG_OPEN_SYSTEM)
		algs |= 1;
	if (auth_alg & AUTH_ALG_SHARED_KEY)
		algs |= 2;
	if (auth_alg & AUTH_ALG_LEAP)
		algs |= 4;
	if (algs == 0)
		algs = 1; /* at least one algorithm should be set */

	wpa_printf(MSG_DEBUG, "%s: auth_alg=0x%x", __FUNCTION__, algs);
	return ipw2100_set_wpa_param(ifname, IPW2100_PARAM_AUTH_ALGS, algs);
}


struct wpa_driver_ops wpa_driver_ipw2100_ops = {
	.get_bssid = wpa_driver_wext_get_bssid,
	.get_ssid = wpa_driver_wext_get_ssid,
	.set_wpa = wpa_driver_ipw2100_set_wpa,
	.set_key = wpa_driver_ipw2100_set_key,
	.events_init = wpa_driver_wext_events_init,
	.events_deinit = wpa_driver_wext_events_deinit,
	.set_countermeasures = wpa_driver_ipw2100_set_countermeasures,
	.set_drop_unencrypted = wpa_driver_ipw2100_set_drop_unencrypted,
	.scan = wpa_driver_wext_scan,
	.get_scan_results = wpa_driver_wext_get_scan_results,
	.deauthenticate = wpa_driver_ipw2100_deauthenticate,
	.disassociate = wpa_driver_ipw2100_disassociate,
	.associate = wpa_driver_ipw2100_associate,
	.set_auth_alg = wpa_driver_ipw2100_set_auth_alg,
};
